==> causaltorch-0.1.0.tar.gz: causaltorch-0.1.0/causaltorch/models.py <==
"""
CausalTorch Models Module
=========================

This module contains pre-built causal neuro-symbolic generative models
for text, image, and video generation.
"""

import torch
import torch.nn as nn
import torch.nn.functional as F

try:
    import pytorch_lightning as pl # type: ignore
    from transformers import GPT2LMHeadModel
except ImportError:
    # Optional dependencies
    pass

from .layers import CausalAttentionLayer, CausalSymbolicLayer, TemporalCausalConv


class CNSG_GPT2(nn.Module):
    """Causal Neuro-Symbolic GPT-2 model for text generation.
    
    This model extends GPT-2 with causal attention to enforce logical
    relationships in generated text.
    
    Args:
        pretrained_model_name (str): Name of pretrained GPT-2 model
        causal_rules (dict): Dictionary of causal rules to enforce
    """
    def __init__(self, pretrained_model_name="gpt2", causal_rules=None):
        super().__init__()
        
        # Load pretrained GPT-2
        self.gpt2 = GPT2LMHeadModel.from_pretrained(pretrained_model_name)
        
        # Add causal attention layer
        self.causal_attn = CausalAttentionLayer(causal_rules or {})
        
        # Set tokenizer (will be initialized by user)
        self.tokenizer = None
    
    def forward(self, input_ids, attention_mask=None, labels=None):
        """Forward pass with causal constraints.
        
        Args:
            input_ids (torch.Tensor): Token IDs
            attention_mask (torch.Tensor, optional): Attention mask
            labels (torch.Tensor, optional): Target token IDs
            
        Returns:
            transformers.modeling_outputs.CausalLMOutputWithCrossAttentions:
                Model outputs with modified attention based on causal rules
        """
        # Run GPT-2 with output_attentions=True to get attention matrices
        outputs = self.gpt2(
            input_ids=input_ids,
            attention_mask=attention_mask,
            labels=labels,
            output_attentions=True
        )
        
        # Get input text
        if self.tokenizer is not None and input_ids is not None:
            input_text = self.tokenizer.decode(input_ids[0])
            
            # Apply causal attention modifications
            if outputs.attentions is not None:
                # For simplicity, we only modify the last layer's attention
                self.causal_attn.tokenizer = self.tokenizer
                modified_attention = self.causal_attn(outputs.attentions[-1], input_text)
                
                # In a full implementation, we would use this modified attention
                # to recompute the final layer's outputs
                
        return outputs
    
    def generate(self, input_ids=None, max_length=None, **kwargs):
        """Generate text with causal constraints.
        
        Args:
            input_ids (torch.Tensor): Input token IDs
            max_length (int, optional): Maximum output length
            **kwargs: Additional generation parameters
            
        Returns:
            torch.Tensor: Generated token IDs
        """
        # For now, we use GPT-2's generation method directly
        # In a full implementation, we would modify the generation
        # algorithm to incorporate causal constraints at each step
        return self.gpt2.generate(
            input_ids=input_ids,
            max_length=max_length,
            **kwargs
        )


class CNSGNet(nn.Module):
    """Causal Neuro-Symbolic Generative Network for image generation.
    
    This model implements a VAE/GAN with causal structure in the latent space.
    
    Args:
        latent_dim (int): Dimension of the latent space
        causal_rules (dict, optional): Dictionary of causal rules
        img_size (int, optional): Size of generated images
    """
    def __init__(self, latent_dim=3, causal_rules=None, img_size=28):
        super().__init__()
        self.latent_dim = latent_dim
        self.img_size = img_size
        
        # Encoder
        self.encoder = nn.Sequential(
            nn.Linear(img_size * img_size, 256),
            nn.ReLU(),
            nn.Linear(256, 128),
            nn.ReLU()
        )
        
        # Mean and variance for VAE
        self.fc_mu = nn.Linear(128, latent_dim)
        self.fc_var = nn.Linear(128, latent_dim)
        
        # Causal layer to enforce relationships in latent space
        self.causal_layer = CausalSymbolicLayer(causal_rules)
        
        # Decoder
        self.decoder = nn.Sequential(
            nn.Linear(latent_dim, 128),
            nn.ReLU(),
            nn.Linear(128, 256),
            nn.ReLU(),
            nn.Linear(256, img_size * img_size),
            nn.Sigmoid()
        )
    
    def encode(self, x):
        """Encode input to latent space.
        
        Args:
            x (torch.Tensor): Input images
            
        Returns:
            tuple: (mu, log_var) parameters of latent distribution
        """
        x = x.view(x.size(0), -1)  # Flatten
        h = self.encoder(x)
        mu = self.fc_mu(h)
        log_var = self.fc_var(h)
        return mu, log_var
    
    def reparameterize(self, mu, log_var):
        """Reparameterization trick for VAE.
        
        Args:
            mu (torch.Tensor): Mean of latent distribution
            log_var (torch.Tensor): Log variance of latent distribution
            
        Returns:
            torch.Tensor: Sampled latent vector
        """
        std = torch.exp(log_var / 2)
        eps = torch.randn_like(std)
        return mu + eps * std
    
    def decode(self, z):
        """Decode latent vector to image.
        
        Args:
            z (torch.Tensor): Latent vector
            
        Returns:
            torch.Tensor: Generated image
        """
        # Apply causal constraints to latent vector
        z = self.causal_layer(z)
        
        # Decode to image
        h = self.decoder(z)
        return h.view(h.size(0), 1, self.img_size, self.img_size)
    
    def forward(self, x):
        """Forward pass through the model.
        
        Args:
            x (torch.Tensor): Input images
            
        Returns:
            tuple: (reconstructed_x, mu, log_var)
        """
        mu, log_var = self.encode(x)
        z = self.reparameterize(mu, log_var)
        x_reconstructed = self.decode(z)
        return x_reconstructed, mu, log_var
    
    def generate(self, rain_intensity=None, num_samples=1):
        """Generate images with causal constraints.
        
        Args:
            rain_intensity (float, optional): Rain intensity value (0-1)
            num_samples (int, optional): Number of images to generate
            
        Returns:
            torch.Tensor: Generated images
        """
        with torch.no_grad():
            # Sample random latent vectors
            z = torch.randn(num_samples, self.latent_dim)
            
            # If rain_intensity is specified, set the rain dimension
            if rain_intensity is not None:
                z[:, 0] = rain_intensity
            
            # Generate images with causal constraints
            images = self.decode(z)
            return images


class CNSG_VideoGenerator(nn.Module):
    """Causal Neuro-Symbolic Video Generator.
    
    This model generates temporally consistent video with causal constraints
    between frames.
    
    Args:
        frame_size (tuple): Height and width of video frames
        latent_dim (int): Dimension of the latent space
        causal_rules (dict, optional): Dictionary of temporal causal rules
    """
    def __init__(self, frame_size=(64, 64), latent_dim=16, causal_rules=None):
        super().__init__()
        self.frame_size = frame_size
        self.latent_dim = latent_dim
        height, width = frame_size
        
        # Frame generator network
        self.generator = nn.Sequential(
            nn.ConvTranspose2d(latent_dim + 3, 64, kernel_size=4, stride=1, padding=0),
            nn.ReLU(),
            nn.ConvTranspose2d(64, 32, kernel_size=4, stride=2, padding=1),
            nn.ReLU(),
            nn.ConvTranspose2d(32, 16, kernel_size=4, stride=2, padding=1),
            nn.ReLU(),
            nn.ConvTranspose2d(16, 3, kernel_size=4, stride=2, padding=1),
            nn.Sigmoid()
        )
        
        # Latent dynamics network (predicts next latent state)
        self.latent_encoder = nn.LSTM(latent_dim, latent_dim, batch_first=True)
        
        # Temporal causal layer
        self.temporal_causal = TemporalCausalConv(3, 3, kernel_size=3, causal_rules=causal_rules or {})
    
    def forward(self, initial_frame, initial_latent, seq_length=24, metadata=None):
        """Generate a video sequence with causal temporal constraints.
        
        Args:
            initial_frame (torch.Tensor): Starting frame [batch, 3, H, W]
            initial_latent (torch.Tensor): Initial latent state [batch, latent_dim]
            seq_length (int): Number of frames to generate
            metadata (dict, optional): Frame metadata for causal rules
            
        Returns:
            torch.Tensor: Generated video sequence [batch, seq_length, 3, H, W]
        """
        batch_size = initial_frame.size(0)
        device = initial_frame.device
        frames = [initial_frame]
        latent = initial_latent
        
        # Generate frames sequentially
        for t in range(seq_length - 1):
            # Get previous frame
            prev_frame = frames[-1]
            
            # Update latent state
            latent_input = latent.unsqueeze(1)  # Add sequence dimension
            latent_output, _ = self.latent_encoder(latent_input)
            latent = latent_output.squeeze(1)  # Remove sequence dimension
            
            # Generate next frame
            gen_input = torch.cat([prev_frame, latent.unsqueeze(-1).unsqueeze(-1).expand(-1, -1, 1, 1)], dim=1)
            next_frame = self.generator(gen_input)
            
            # Apply temporal causal effects
            if metadata is not None:
                # Get metadata for current frame
                frame_metadata = {k: v[t] if isinstance(v, list) else v for k, v in metadata.items()}
                next_frame = self.temporal_causal(next_frame.unsqueeze(2), frame_metadata).squeeze(2)
            
            frames.append(next_frame)
        
        # Stack frames to make video
        video = torch.stack(frames, dim=1)  # [batch, seq_length, 3, H, W]
        return video
    
    def generate_battle_scene(self, num_frames=24):
        """Generate a battle scene with horses, arrows, and causal effects.
        
        Args:
            num_frames (int): Number of frames to generate
            
        Returns:
            torch.Tensor: Generated battle video
        """
        # Create initial inputs
        batch_size = 1
        initial_frame = torch.randn(batch_size, 3, self.frame_size[0], self.frame_size[1])
        initial_latent = torch.zeros(batch_size, self.latent_dim)
        
        # Set up metadata with causal events
        metadata = {
            # Hoof contacts ground at specific frames
            "hoof_contact": [1.0 if i % 6 == 0 else 0.0 for i in range(num_frames)],
            
            # Arrow hits at frame 10
            "arrow_hit": [1.0 if i == 10 else 0.0 for i in range(num_frames)]
        }
        
        # Generate video
        return self.forward(initial_frame, initial_latent, seq_length=num_frames, metadata=metadata)

==> causaltorch-2.0.0.tar.gz: causaltorch-2.0.0/causaltorch/models.py <==
"""
CausalTorch Models Module
=========================

This module contains pre-built causal neuro-symbolic generative models
for text, image, and video generation.
"""

import torch
import torch.nn as nn
import torch.nn.functional as F

try:
    import pytorch_lightning as pl # type: ignore
    from transformers import GPT2LMHeadModel # type: ignore
except ImportError:
    # Optional dependencies
    pass

from .layers import CausalAttentionLayer, CausalSymbolicLayer, TemporalCausalConv


class CNSG_GPT2(nn.Module):
    """Causal Neuro-Symbolic GPT-2 model for text generation.
    
    This model extends GPT-2 with causal attention to enforce logical
    relationships in generated text.
    
    Args:
        pretrained_model_name (str): Name of pretrained GPT-2 model
        causal_rules (dict): Dictionary of causal rules to enforce
    """
    def __init__(self, pretrained_model_name="gpt2", causal_rules=None):
        super().__init__()
        
        # Load pretrained GPT-2
        self.gpt2 = GPT2LMHeadModel.from_pretrained(pretrained_model_name)
        
        # Add causal attention layer
        self.causal_attn = CausalAttentionLayer(causal_rules or {})
        
        # Set tokenizer (will be initialized by user)
        self.tokenizer = None
    
    def forward(self, input_ids, attention_mask=None, labels=None):
        """Forward pass with causal constraints.
        
        Args:
            input_ids (torch.Tensor): Token IDs
            attention_mask (torch.Tensor, optional): Attention mask
            labels (torch.Tensor, optional): Target token IDs
            
        Returns:
            transformers.modeling_outputs.CausalLMOutputWithCrossAttentions:
                Model outputs with modified attention based on causal rules
        """
        # Run GPT-2 with output_attentions=True to get attention matrices
        outputs = self.gpt2(
            input_ids=input_ids,
            attention_mask=attention_mask,
            labels=labels,
            output_attentions=True
        )
        
        # Get input text
        if self.tokenizer is not None and input_ids is not None:
            input_text = self.tokenizer.decode(input_ids[0])
            
            # Apply causal attention modifications
            if outputs.attentions is not None:
                # For simplicity, we only modify the last layer's attention
                self.causal_attn.tokenizer = self.tokenizer
                modified_attention = self.causal_attn(outputs.attentions[-1], input_text)
                
                # In a full implementation, we would use this modified attention
                # to recompute the final layer's outputs
                
        return outputs
    
    def generate(self, input_ids=None, max_length=None, **kwargs):
        """Generate text with causal constraints.
        
        Args:
            input_ids (torch.Tensor): Input token IDs
            max_length (int, optional): Maximum output length
            **kwargs: Additional generation parameters
            
        Returns:
            torch.Tensor: Generated token IDs
        """
        # For now, we use GPT-2's generation method directly
        # In a full implementation, we would modify the generation
        # algorithm to incorporate causal constraints at each step
        return self.gpt2.generate(
            input_ids=input_ids,
            max_length=max_length,
            **kwargs
        )


class CNSGNet(nn.Module):
    """Causal Neuro-Symbolic Generative Network for image generation.
    
    This model implements a VAE/GAN with causal structure in the latent space.
    
    Args:
        latent_dim (int): Dimension of the latent space
        causal_rules (dict, optional): Dictionary of causal rules
        img_size (int, optional): Size of generated images
    """
    def __init__(self, latent_dim=3, causal_rules=None, img_size=28):
        super().__init__()
        self.latent_dim = latent_dim
        self.img_size = img_size
        
        # Encoder
        self.encoder = nn.Sequential(
            nn.Linear(img_size * img_size, 256),
            nn.ReLU(),
            nn.Linear(256, 128),
            nn.ReLU()
        )
        
        # Mean and variance for VAE
        self.fc_mu = nn.Linear(128, latent_dim)
        self.fc_var = nn.Linear(128, latent_dim)
        
        # Causal layer to enforce relationships in latent space
        self.causal_layer = CausalSymbolicLayer(causal_rules)
        
        # Decoder
        self.decoder = nn.Sequential(
            nn.Linear(latent_dim, 128),
            nn.ReLU(),
            nn.Linear(128, 256),
            nn.ReLU(),
            nn.Linear(256, img_size * img_size),
            nn.Sigmoid()
        )
    
    def encode(self, x):
        """Encode input to latent space.
        
        Args:
            x (torch.Tensor): Input images
            
        Returns:
            tuple: (mu, log_var) parameters of latent distribution
        """
        x = x.view(x.size(0), -1)  # Flatten
        h = self.encoder(x)
        mu = self.fc_mu(h)
        log_var = self.fc_var(h)
        return mu, log_var
    
    def reparameterize(self, mu, log_var):
        """Reparameterization trick for VAE.
        
        Args:
            mu (torch.Tensor): Mean of latent distribution
            log_var (torch.Tensor): Log variance of latent distribution
            
        Returns:
            torch.Tensor: Sampled latent vector
        """
        std = torch.exp(log_var / 2)
        eps = torch.randn_like(std)
        return mu + eps * std
    
    def decode(self, z):
        """Decode latent vector to image.
        
        Args:
            z (torch.Tensor): Latent vector
            
        Returns:
            torch.Tensor: Generated image
        """
        # Apply causal constraints to latent vector
        z = self.causal_layer(z)
        
        # Decode to image
        h = self.decoder(z)
        return h.view(h.size(0), 1, self.img_size, self.img_size)
    
    def forward(self, x):
        """Forward pass through the model.
        
        Args:
            x (torch.Tensor): Input images
            
        Returns:
            tuple: (reconstructed_x, mu, log_var)
        """
        mu, log_var = self.encode(x)
        z = self.reparameterize(mu, log_var)
        x_reconstructed = self.decode(z)
        return x_reconstructed, mu, log_var
    
    def generate(self, rain_intensity=None, num_samples=1):
        """Generate images with causal constraints.
        
        Args:
            rain_intensity (float, optional): Rain intensity value (0-1)
            num_samples (int, optional): Number of images to generate
            
        Returns:
            torch.Tensor: Generated images
        """
        with torch.no_grad():
            # Sample random latent vectors
            z = torch.randn(num_samples, self.latent_dim)
            
            # If rain_intensity is specified, set the rain dimension
            if rain_intensity is not None:
                z[:, 0] = rain_intensity
            
            # Generate images with causal constraints
            images = self.decode(z)
            return images


class CNSG_VideoGenerator(nn.Module):
    """Causal Neuro-Symbolic Video Generator.
    
    This model generates temporally consistent video with causal constraints
    between frames.
    
    Args:
        frame_size (tuple): Height and width of video frames
        latent_dim (int): Dimension of the latent space
        causal_rules (dict, optional): Dictionary of temporal causal rules
    """
    def __init__(self, frame_size=(64, 64), latent_dim=16, causal_rules=None):
        super().__init__()
        self.frame_size = frame_size
        self.latent_dim = latent_dim
        height, width = frame_size
        
        # Frame generator network
        self.generator = nn.Sequential(
            nn.ConvTranspose2d(latent_dim + 3, 64, kernel_size=4, stride=1, padding=0),
            nn.ReLU(),
            nn.ConvTranspose2d(64, 32, kernel_size=4, stride=2, padding=1),
            nn.ReLU(),
            nn.ConvTranspose2d(32, 16, kernel_size=4, stride=2, padding=1),
            nn.ReLU(),
            nn.ConvTranspose2d(16, 3, kernel_size=4, stride=2, padding=1),
            nn.Sigmoid()
        )
        
        # Latent dynamics network (predicts next latent state)
        self.latent_encoder = nn.LSTM(latent_dim, latent_dim, batch_first=True)
        
        # Temporal causal layer
        self.temporal_causal = TemporalCausalConv(3, 3, kernel_size=3, causal_rules=causal_rules or {})
    
    def forward(self, initial_frame, initial_latent, seq_length=24, metadata=None):
        """Generate a video sequence with causal temporal constraints.
        
        Args:
            initial_frame (torch.Tensor): Starting frame [batch, 3, H, W]
            initial_latent (torch.Tensor): Initial latent state [batch, latent_dim]
            seq_length (int): Number of frames to generate
            metadata (dict, optional): Frame metadata for causal rules
            
        Returns:
            torch.Tensor: Generated video sequence [batch, seq_length, 3, H, W]
        """
        batch_size = initial_frame.size(0)
        device = initial_frame.device
        frames = [initial_frame]
        latent = initial_latent
        
        # Generate frames sequentially
        for t in range(seq_length - 1):
            # Get previous frame
            prev_frame = frames[-1]
            
            # Update latent state
            latent_input = latent.unsqueeze(1)  # Add sequence dimension
            latent_output, _ = self.latent_encoder(latent_input)
            latent = latent_output.squeeze(1)  # Remove sequence dimension
            
            # Generate next frame
            gen_input = torch.cat([prev_frame, latent.unsqueeze(-1).unsqueeze(-1).expand(-1, -1, 1, 1)], dim=1)
            next_frame = self.generator(gen_input)
            
            # Apply temporal causal effects
            if metadata is not None:
                # Get metadata for current frame
                frame_metadata = {k: v[t] if isinstance(v, list) else v for k, v in metadata.items()}
                next_frame = self.temporal_causal(next_frame.unsqueeze(2), frame_metadata).squeeze(2)
            
            frames.append(next_frame)
        
        # Stack frames to make video
        video = torch.stack(frames, dim=1)  # [batch, seq_length, 3, H, W]
        return video
    
    def generate_battle_scene(self, num_frames=24):
        """Generate a battle scene with horses, arrows, and causal effects.
        
        Args:
            num_frames (int): Number of frames to generate
            
        Returns:
            torch.Tensor: Generated battle video
        """
        # Create initial inputs
        batch_size = 1
        initial_frame = torch.randn(batch_size, 3, self.frame_size[0], self.frame_size[1])
        initial_latent = torch.zeros(batch_size, self.latent_dim)
        
        # Set up metadata with causal events
        metadata = {
            # Hoof contacts ground at specific frames
            "hoof_contact": [1.0 if i % 6 == 0 else 0.0 for i in range(num_frames)],
            
            # Arrow hits at frame 10
            "arrow_hit": [1.0 if i == 10 else 0.0 for i in range(num_frames)]
        }
        
        # Generate video
        return self.forward(initial_frame, initial_latent, seq_length=num_frames, metadata=metadata)
